import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv2D, Dense, Flatten, Dropout, BatchNormalization, Activation, MaxPooling2D
from tensorflow.keras.constraints import max_norm
from tensorflow.keras.callbacks import CSVLogger, ModelCheckpoint, ReduceLROnPlateau, EarlyStopping

import numpy as np
import os
import time
from sklearn.metrics import classification_report, f1_score
from min2net.utils import TimeHistory, compute_class_weight

class DeepConvNet:
    def __init__(self,
                input_shape=(1,20,400),
                num_class=2,
                loss='sparse_categorical_crossentropy',
                epochs=200,
                batch_size=100,
                optimizer = Adam(beta_1=0.9, beta_2=0.999, epsilon=1e-08),
                lr=0.01,
                min_lr=0.01,
                factor=0.25,
                patience=10,
                es_patience=20,
                verbose=1,
                log_path='logs',
                model_name='DeepConvNet',
                **kwargs):
        self.input_shape = input_shape
        self.num_class = num_class
        self.loss = loss
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.optimizer.lr = lr
        self.lr = lr
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.es_patience = es_patience
        self.verbose = verbose
        self.log_path = log_path
        self.model_name = model_name
        self.weights_dir = log_path+'/'+model_name+'_out_weights.h5'
        self.csv_dir = log_path+'/'+model_name+'_out_log.log'
        self.time_log = log_path+'/'+model_name+'_time_log.csv'

        # use **kwargs to set the new value of below args.
        self.kernLength = 125
        self.F1 = 8
        self.D = 2
        self.F2 = int(self.F1*self.D)
        self.norm_rate = 0.25
        self.dropout_rate = 0.5
        self.f1_average = 'binary' if self.num_class == 2 else 'macro'
        self.data_format = 'channels_first'
        self.shuffle = False
        self.metrics = 'accuracy'
        self.monitor = 'val_loss'
        self.mode = 'min'
        self.save_best_only = True
        self.save_weight_only = True
        self.seed = 1234
        self.class_balancing = False
        self.class_weight = None

        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
            
        if self.data_format == 'channels_first':
            self.Chans = self.input_shape[1]
            self.Samples = self.input_shape[2]
        else:
            self.Chans = self.input_shape[0]
            self.Samples = self.input_shape[1]

        np.random.seed(self.seed)
        tf.random.set_seed(self.seed)
        K.set_image_data_format(self.data_format)
        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)

    def build(self):
        """ Keras implementation of the Deep Convolutional Network as described in
        Schirrmeister et. al. (2017), Human Brain Mapping.

        This implementation assumes the input is a 2-second EEG signal sampled at
        128Hz, as opposed to signals sampled at 250Hz as described in the original
        paper. We also perform temporal convolutions of length (1, 5) as opposed
        to (1, 10) due to this sampling rate difference.

        Note that we use the max_norm constraint on all convolutional layers, as
        well as the classification layer. We also change the defaults for the
        BatchNormalization layer. We used this based on a personal communication
        with the original authors.

                          ours        original paper
        pool_size        1, 2        1, 3
        strides          1, 2        1, 3
        conv filters     1, 5        1, 10

        Note that this implementation has not been verified by the original
        authors.

        """

        # start the model
        input_main   = Input(self.input_shape)
        block1       = Conv2D(25, (1, 5),
                              input_shape=(self.input_shape),
                              kernel_constraint=max_norm(2., axis=(0, 1, 2)))(input_main)
        block1       = Conv2D(25, (self.Chans, 1),
                              kernel_constraint=max_norm(2., axis=(0, 1, 2)))(block1)
        block1       = BatchNormalization(axis=1, epsilon=1e-05, momentum=0.1)(block1)
        block1       = Activation('elu')(block1)
        block1       = MaxPooling2D(pool_size=(1, 2), strides=(1, 2))(block1)
        block1       = Dropout(self.dropout_rate)(block1)

        block2       = Conv2D(50, (1, 5),
                              kernel_constraint=max_norm(2., axis=(0, 1, 2)))(block1)
        block2       = BatchNormalization(axis=1, epsilon=1e-05, momentum=0.1)(block2)
        block2       = Activation('elu')(block2)
        block2       = MaxPooling2D(pool_size=(1, 2), strides=(1, 2))(block2)
        block2       = Dropout(self.dropout_rate)(block2)

        block3       = Conv2D(100, (1, 5),
                              kernel_constraint=max_norm(2., axis=(0, 1, 2)))(block2)
        block3       = BatchNormalization(axis=1, epsilon=1e-05, momentum=0.1)(block3)
        block3       = Activation('elu')(block3)
        block3       = MaxPooling2D(pool_size=(1, 2), strides=(1, 2))(block3)
        block3       = Dropout(self.dropout_rate)(block3)

        block4       = Conv2D(200, (1, 5),
                              kernel_constraint=max_norm(2., axis=(0, 1, 2)))(block3)
        block4       = BatchNormalization(axis=1, epsilon=1e-05, momentum=0.1)(block4)
        block4       = Activation('elu')(block4)
        block4       = MaxPooling2D(pool_size=(1, 2), strides=(1, 2))(block4)
        block4       = Dropout(self.dropout_rate)(block4)

        flatten      = Flatten()(block4)

        dense        = Dense(self.num_class, kernel_constraint = max_norm(0.5))(flatten)
        softmax      = Activation('softmax')(dense)

        return Model(inputs=input_main, outputs=softmax)

    def fit(self, X_train, y_train, X_val, y_val):

        if X_train.ndim != 4:
            raise Exception('ValueError: `X_train` is incompatible: expected ndim=4, found ndim='+str(X_train.ndim))
        elif X_val.ndim != 4:
            raise Exception('ValueError: `X_val` is incompatible: expected ndim=4, found ndim='+str(X_val.ndim))

        self.input_shape = X_train.shape[1:]
        if self.data_format == 'channels_first':
            self.Chans = self.input_shape[1]
            self.Samples = self.input_shape[2]
        else:
            self.Chans = self.input_shape[0]
            self.Samples = self.input_shape[1]
        
        csv_logger = CSVLogger(self.csv_dir)
        time_callback = TimeHistory(self.time_log)
        checkpointer = ModelCheckpoint(monitor=self.monitor, filepath=self.weights_dir, verbose=self.verbose, 
                                       save_best_only=self.save_best_only, save_weight_only=self.save_weight_only)
        reduce_lr = ReduceLROnPlateau(monitor=self.monitor, patience=self.patience, factor=self.factor, mode=self.mode, 
                                      verbose=self.verbose, min_lr=self.min_lr)
        es = EarlyStopping(monitor=self.monitor, mode=self.mode, verbose=self.verbose, patience=self.es_patience)

        model = self.build()
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
        model.summary()
        
        if self.class_balancing: # compute_class_weight if class_balancing is True
            self.class_weight = compute_class_weight(y_train)
        else:
            self.class_weight = None
            
        model.fit(X_train, y_train,
                  batch_size=self.batch_size, shuffle=self.shuffle,
                  epochs=self.epochs, validation_data=(X_val, y_val), class_weight=self.class_weight,
                  callbacks=[checkpointer,csv_logger,reduce_lr,es, time_callback])

    def predict(self, X_test, y_test):

        if X_test.ndim != 4:
            raise Exception('ValueError: `X_test` is incompatible: expected ndim=4, found ndim='+str(X_test.ndim))

        model = self.build()
        model.load_weights(self.weights_dir)
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)

        start = time.time()
        y_pred = model.predict(X_test)
        end = time.time()
        loss, accuracy = model.evaluate(x=X_test, y=y_test, batch_size=self.batch_size, verbose=self.verbose)
        y_pred_argm = np.argmax(y_pred, axis=1)
        print(classification_report(y_test, y_pred_argm))
        print("F1-score is computed based on {}".format(self.f1_average))
        f1 = f1_score(y_test, y_pred_argm, average=self.f1_average)
        evaluation = {'loss': loss, 
                      'accuracy': accuracy, 
                      'f1-score': f1, 
                      'prediction_time': end-start}
        Y = {'y_true': y_test, 
             'y_pred': y_pred_argm}
        return Y, evaluation



import tensorflow as tf
import tensorflow_addons as tfa
import tensorflow.keras.backend as K


def mean_squared_error(y_true, y_pred):
	""" loss function computing MSE of non-blank(!=0) in y_true
		Args:
			y_true(tftensor): true label
			y_pred(tftensor): predicted label
		return:
			MSE reconstruction error for loss computing
	"""
	loss = K.switch(K.equal(y_true, tf.constant(0.)),tf.zeros(K.shape(y_true)),K.square(y_pred - y_true))
	return K.mean(loss, axis=-1)

def triplet_loss(margin = 1.0):
    def inner_triplet_loss_objective(y_true, y_pred):
        labels = y_true
        embeddings = y_pred
        return tfa.losses.triplet_semihard_loss(y_true=labels, y_pred=embeddings,margin=margin)
    return inner_triplet_loss_objective

def SparseCategoricalCrossentropy(class_weight = None):
    """[SparseCategoricalCrossentropy]

    Args:
        class_weight ([dict], optional): dict of class_weight
        class_weight = {0: 0.3,
                        1: 0.7}
        Defaults to None.
    """
    def inner_sparse_categorical_crossentropy(y_true, y_pred):
        scce = tf.keras.losses.SparseCategoricalCrossentropy()
        if class_weight:
            keys_tensor = tf.cast(tf.constant(list(class_weight.keys())), dtype=tf.int32)
            vals_tensor = tf.constant(list(class_weight.values()), tf.float32)
            input_tensor = tf.cast(y_true, dtype=tf.int32)
            init = tf.lookup.KeyValueTensorInitializer(keys_tensor, vals_tensor)
            table = tf.lookup.StaticHashTable(init, default_value=-1)
            sample_weight = table.lookup(input_tensor)
        else:
            sample_weight = None
        return scce(y_true, y_pred, sample_weight)
    return inner_sparse_categorical_crossentropy


import numpy as np
from scipy import linalg 
from sklearn.feature_selection import SelectKBest, mutual_info_classif
from min2net.utils import butter_bandpass_filter
# Note that we modify CSP function using MNE-Python package (version 0.20).
# Reference: 
# - https://github.com/mne-tools/mne-python
# - A. Gramfort, M. Luessi, E. Larson, D. Engemann, D. Strohmeier, C. Brodbeck, R. Goj, M. Jas, T. Brooks, L. Parkkonen, and M. Ha ̈ma ̈la ̈inen, “Meg and eeg data analysis with mne-python,” Frontiers in Neuroscience, vol. 7, p. 267, 2013.
# Sources:
# - Common Spatial Pattern revisited by Riemannian geometry
# - Model based generalization analysis of common spatial pattern in brain computer interfaces

class FBCSP():
    '''
    Two- or multi-class FBCSP
    '''
    def __init__(self, 
                bands, 
                smp_freq, 
                num_class=2, 
                order=5, 
                n_components=4, 
                n_features=8):
        self.bands = bands
        self.smp_freq = smp_freq
        self.n_components = n_components
        self.n_features = n_features
        self.cov_type = 'epoch'
        self.order = order
        self.n_classes = num_class
        self.classes = None
        if self.n_classes == 2:
            self.component_order = 'alternate'
        else:
            self.component_order = 'mutual_info'
        self.norm_trace = False
        self.spatial_transform = [None for i in range(len(bands))]
        self.selector = SelectKBest(score_func=mutual_info_classif, k=self.n_features)
        
    def __covariance(self, X):
#         covariance = np.dot(X, X.T)/X.shape[0]
#         return covariance
        return np.dot(X,X.T)/np.trace(np.dot(X,X.T))
   
    # spatialFilter returns the spatial filter for mean covariance matrices of two classes 
    def __get_spatial_filter(self, covs, sample_weights):
        eigen_vectors, eigen_values = self.__decompose_covs(covs, sample_weights)
        # CSP requires the eigenvalues and eigenvector be sorted in descending order
        sorted_index = self.__order_components(covs, sample_weights, eigen_vectors, eigen_values) 
        eigen_vectors = eigen_vectors[:, sorted_index]
        sp_filters = eigen_vectors.T
        pick_filters = sp_filters[:self.n_components]
        return pick_filters, sp_filters
    
    def __decompose_covs(self, covs, sample_weights):
        n_classes = len(covs)
        if n_classes == 2:
            eigen_values, eigen_vectors = linalg.eigh(covs[0], covs.sum(0))
        else:
            # The multiclass case is adapted from
            # http://github.com/alexandrebarachant/pyRiemann
            eigen_vectors, D = self.__ajd_pham(covs)
            eigen_vectors = self.__normalize_eigenvectors(eigen_vectors.T, covs, sample_weights)
            eigen_values = None
        return eigen_vectors, eigen_values
    
    def __calculate_covariance_matrices(self, data, y_class):
        '''The data is in the form of samples x channels x sampled_time_points'''
        if len(data.shape) != 3:
            raise Exception('Dimension is not match!')
        n_samples, n_channels, n_points = data.shape    
        
        if self.cov_type == 'concat':
            cov_estimator = self.__concat_cov
        elif self.cov_type == 'epoch':
            cov_estimator = self.__epoch_cov
            
        covs = []
        sample_weights = []
        self.classes = np.unique(y_class)
        n_classes = len(self.classes)
        for id_class in self.classes:
            cov, weight = cov_estimator(data[y_class == id_class])
            if self.norm_trace:
                cov /= np.trace(cov)
            covs.append(cov)
            sample_weights.append(weight)
        return np.stack(covs), np.array(sample_weights)
    
    def __concat_cov(self, X_class):
        '''The data is in the form of samples x channels x sampled_time_points'''
        '''Concatenate epochs before computing the covariance.'''
        n_samples, n_channels, n_points = X_class.shape
        X_class = np.transpose(X_class, [1, 0, 2])
        X_class = X_class.reshape(n_channels, -1)
        # The covariace with norm trace is used to provide better results
        cov = self.__covariance(X_class)
        # cov = np.cov(X_class) 
        weight = X_class.shape[0]
        return cov, weight

    def __epoch_cov(self, X_class):
        '''The data is in the form of samples x channels x sampled_time_points'''
        '''Mean of per-epoch covariances.'''
        cov = sum(self.__covariance(data) for data in X_class)
        cov /= len(X_class)
        weight = len(X_class)
        return cov, weight
    
    def __compute_mutual_info(self, covs, sample_weights, eigen_vectors):
        class_probas = sample_weights / sample_weights.sum()

        mutual_info = []
        for jj in range(eigen_vectors.shape[1]):
            aa, bb = 0, 0
            for (cov, prob) in zip(covs, class_probas):
                tmp = np.dot(np.dot(eigen_vectors[:, jj].T, cov),
                             eigen_vectors[:, jj])
                aa += prob * np.log(np.sqrt(tmp))
                bb += prob * (tmp ** 2 - 1)
            mi = - (aa + (3.0 / 16) * (bb ** 2))
            mutual_info.append(mi)

        return mutual_info

    def __normalize_eigenvectors(self, eigen_vectors, covs, sample_weights):
        # Here we apply an euclidean mean. See pyRiemann for other metrics
        mean_cov = np.average(covs, axis=0, weights=sample_weights)

        for ii in range(eigen_vectors.shape[1]):
            tmp = np.dot(np.dot(eigen_vectors[:, ii].T, mean_cov),
                         eigen_vectors[:, ii])
            eigen_vectors[:, ii] /= np.sqrt(tmp)
        return eigen_vectors
    
    def __order_components(self, covs, sample_weights, eigen_vectors, eigen_values):
        n_classes = len(self.classes)
        ix = []
        if self.component_order == 'mutual_info' and n_classes > 2:
            mutual_info = self.__compute_mutual_info(covs, sample_weights, eigen_vectors)
            ix = np.argsort(mutual_info)[::-1]
        elif self.component_order == 'mutual_info' and n_classes == 2:
            ix = np.argsort(np.abs(eigen_values - 0.5))[::-1]
        elif self.component_order == 'alternate' and n_classes == 2:
            i = np.argsort(eigen_values)
            ix = np.empty_like(i)
            ix[1::2] = i[:len(i) // 2]
            ix[0::2] = i[len(i) // 2:][::-1]
        return ix
    
    def __ajd_pham(self, X, eps=1e-6, max_iter=15):
        '''Approximate joint diagonalization based on Pham's algorithm.
        This is a direct implementation of the PHAM's AJD algorithm [1].
        Parameters
        ----------
        X : ndarray, shape (n_epochs, n_channels, n_channels)
            A set of covariance matrices to diagonalize.
        eps : float, default 1e-6
            The tolerance for stopping criterion.
        max_iter : int, default 1000
            The maximum number of iteration to reach convergence.
        Returns
        -------
        V : ndarray, shape (n_channels, n_channels)
            The diagonalizer.
        D : ndarray, shape (n_epochs, n_channels, n_channels)
            The set of quasi diagonal matrices.
        References
        ----------
        .. [1] Pham, Dinh Tuan. 'Joint approximate diagonalization of positive
               definite Hermitian matrices.' SIAM Journal on Matrix Analysis and
               Applications 22, no. 4 (2001): 1136-1152.
        '''
        # Adapted from http://github.com/alexandrebarachant/pyRiemann
        n_epochs = X.shape[0]

        # Reshape input matrix
        A = np.concatenate(X, axis=0).T

        # Init variables
        n_times, n_m = A.shape
        V = np.eye(n_times)
        epsilon = n_times * (n_times - 1) * eps

        for it in range(max_iter):
            decr = 0
            for ii in range(1, n_times):
                for jj in range(ii):
                    Ii = np.arange(ii, n_m, n_times)
                    Ij = np.arange(jj, n_m, n_times)

                    c1 = A[ii, Ii]
                    c2 = A[jj, Ij]

                    g12 = np.mean(A[ii, Ij] / c1)
                    g21 = np.mean(A[ii, Ij] / c2)

                    omega21 = np.mean(c1 / c2)
                    omega12 = np.mean(c2 / c1)
                    omega = np.sqrt(omega12 * omega21)

                    tmp = np.sqrt(omega21 / omega12)
                    tmp1 = (tmp * g12 + g21) / (omega + 1)
                    tmp2 = (tmp * g12 - g21) / max(omega - 1, 1e-9)

                    h12 = tmp1 + tmp2
                    h21 = np.conj((tmp1 - tmp2) / tmp)

                    decr += n_epochs * (g12 * np.conj(h12) + g21 * h21) / 2.0

                    tmp = 1 + 1.j * 0.5 * np.imag(h12 * h21)
                    tmp = np.real(tmp + np.sqrt(tmp ** 2 - h12 * h21))
                    tau = np.array([[1, -h12 / tmp], [-h21 / tmp, 1]])

                    A[[ii, jj], :] = np.dot(tau, A[[ii, jj], :])
                    tmp = np.c_[A[:, Ii], A[:, Ij]]
                    tmp = np.reshape(tmp, (n_times * n_epochs, 2), order='F')
                    tmp = np.dot(tmp, tau.T)

                    tmp = np.reshape(tmp, (n_times, n_epochs * 2), order='F')
                    A[:, Ii] = tmp[:, :n_epochs]
                    A[:, Ij] = tmp[:, n_epochs:]
                    V[[ii, jj], :] = np.dot(tau, V[[ii, jj], :])
            if decr < epsilon:
                break
        D = np.reshape(A, (n_times, -1, n_times)).transpose(1, 0, 2)
        return V, D

    def __get_log_var_feats(self, spatial_filt, data):
        data_dot = np.dot(spatial_filt, data) # spatially filtered signals 
        data_var = np.var(data_dot, axis=1)
        # data_var = (data_dot**2).mean(axis=1)
        # We use log(var) instead of averaging
        data_log = np.log(data_var)
        return data_log
    
    def fit_transform(self, X, y):  
        if len(X.shape) != 3:
            raise Exception('Dimension is not match!')
                  
        n_samples = X.shape[0]
        X_transformed_var = np.zeros((len(self.bands), n_samples, self.n_components))

        for id_band, freq_band in enumerate(self.bands):
            # Compute band-pass filter of EEG signals
            X_filtered = butter_bandpass_filter(X, freq_band[0], freq_band[1], self.smp_freq, self.order)
            
            # Calculating covariance only on training set
            covs, sample_weights =  self.__calculate_covariance_matrices(X_filtered, y)
            spf_sel, spf_org = self.__get_spatial_filter(covs, sample_weights)
            self.spatial_transform[id_band] = spf_sel
                  
            # Calculate the variance of spatially filtered signals and then compute the logarithm 
            for sample in range(X_filtered.shape[0]):
                X_transformed_var[id_band, sample] = self.__get_log_var_feats(self.spatial_transform[id_band], X_filtered[sample,:,:])
        
        X_transformed_var = np.swapaxes(X_transformed_var, 0, 1)
        X_transformed_var = X_transformed_var.reshape(n_samples, -1)

        # select k best
        X_fbcsp = self.selector.fit_transform(X_transformed_var, y) 
        return X_fbcsp

    def transform(self, X):
        if len(X.shape) != 3:
            raise Exception('Dimension is not match!')
        
        n_samples = X.shape[0]
        X_transformed_var = np.zeros((len(self.bands), n_samples, self.n_components))

        for id_band, freq_band in enumerate(self.bands):
            X_filtered = butter_bandpass_filter(X, freq_band[0], freq_band[1], self.smp_freq, self.order)
            
            # Calculate the variance of spatially filtered signals and then compute the logarithm 
            for sample_te in range(X_filtered.shape[0]):
                X_transformed_var[id_band, sample_te] = self.__get_log_var_feats(self.spatial_transform[id_band], X_filtered[sample_te,:,:]) 

        X_transformed_var = np.swapaxes(X_transformed_var, 0, 1)
        X_transformed_var = X_transformed_var.reshape(n_samples, -1)
        # select k best
        X_fbcsp = self.selector.transform(X_transformed_var)  
        return X_fbcsp


import numpy as np
import scipy.io as sio
from min2net.utils import resampling
from min2net.preprocessing.config import CONSTANT
CONSTANT = CONSTANT['BCIC2a']
n_chs = CONSTANT['n_chs']
n_trials = 2*CONSTANT['n_trials']
window_len = CONSTANT['trial_len']*CONSTANT['orig_smp_freq']
orig_chs = CONSTANT['orig_chs']
trial_len = CONSTANT['trial_len'] 
orig_smp_freq = CONSTANT['orig_smp_freq']

def read_raw(PATH, subject, training, num_class, id_chosen_chs):
    data = np.zeros((n_trials, n_chs, window_len))
    label = np.zeros(n_trials)

    NO_valid_trial = 0
    if training:
        mat = sio.loadmat(PATH+'/A0'+str(subject)+'T.mat')['data']
    else:
        mat = sio.loadmat(PATH+'/A0'+str(subject)+'E.mat')['data']
    for ii in range(0,mat.size):
        mat_1 = mat[0,ii]
        mat_2 = [mat_1[0,0]]
        mat_info = mat_2[0]
        _X = mat_info[0]
        _trial = mat_info[1]
        _y = mat_info[2]
        _fs = mat_info[3]
        _classes = mat_info[4]
        _artifacts = mat_info[5]
        _gender = mat_info[6]
        _age = mat_info[7]
        for trial in range(0,_trial.size):
            # num_class = 2: picked only class 1 (left hand) and class 2 (right hand) for our propose
            if(_y[trial][0] <= num_class): 
                data[NO_valid_trial,:,:] = np.transpose(_X[int(_trial[trial]):(int(_trial[trial])+window_len),1:21]) # selected merely motor cortices region
                label[NO_valid_trial] = int(_y[trial])
                NO_valid_trial +=1
    data = data[0:NO_valid_trial, id_chosen_chs, :]
    label = label[0:NO_valid_trial]-1 # -1 to adjust the values of class to be in range 0 and 1
    return data, label

def chanel_selection(sel_chs): 
    chs_id = []
    for name_ch in sel_chs:
        ch_id = np.where(np.array(orig_chs) == name_ch)[0][0]
        chs_id.append(ch_id)
        print('chosen_channel:', name_ch, '---', 'Index_is:', ch_id)
    return chs_id
        
def load_crop_data(PATH, subject, start, stop, new_smp_freq, num_class, id_chosen_chs):
    start_time = int(start*new_smp_freq) # 2*
    stop_time = int(stop*new_smp_freq) # 6*
    X_train, y_tr = read_raw(PATH=PATH, subject=subject,
                             training=True, num_class=num_class, id_chosen_chs=id_chosen_chs)
    X_test, y_te = read_raw(PATH=PATH, subject=subject,
                            training=False, num_class=num_class, id_chosen_chs=id_chosen_chs)
    if new_smp_freq < orig_smp_freq:
        X_train = resampling(X_train, new_smp_freq, trial_len)
        X_test = resampling(X_test, new_smp_freq, trial_len)
    X_train = X_train[:,:,start_time:stop_time]
    X_test = X_test[:,:,start_time:stop_time]
    print("Verify dimension training {} and testing {}".format(X_train.shape, X_test.shape)) 
    return X_train, y_tr, X_test, y_te 


import numpy as np
import csv
from scipy import signal
from scipy.signal import butter, filtfilt
import wget
import os
import time
import tensorflow as tf
from sklearn.utils import class_weight
from scipy.interpolate import CubicSpline 
from scipy import ndimage
import argparse

# lib path
PATH = os.path.dirname(os.path.realpath(__file__))

def load_raw(dataset):
    # folder_name = str(PATH)+'/datasets'
    folder_name = 'datasets'
    if dataset == 'OpenBMI':
        try:
            num_subjects = 54
            sessions = [1, 2]
            save_path = folder_name + '/' + dataset + '/raw'
            if save_path is not None:
                if not os.path.exists(save_path):
                    os.makedirs(save_path)
            for session in sessions:
                for person in range(1, num_subjects+1):
                    file_name = '/sess{:02d}_subj{:02d}_EEG_MI.mat'.format(session,person)
                    if os.path.exists(save_path+file_name):
                        os.remove(save_path+file_name) # if exist, remove file
                    print('\n===Download is being processed on session: {} subject: {}==='.format(session, person))
                    url = 'ftp://parrot.genomics.cn/gigadb/pub/10.5524/100001_101000/100542/session{}/s{}{}'.format(session, person, file_name)
                    print('save to: '+save_path+file_name)
                    wget.download(url,  save_path+file_name)
            print('\nDone!')
        except:
            raise Exception('Path Error: file does not exist, please direccly download at http://gigadb.org/dataset/100542')
    elif dataset == 'BCIC2a':
        try:
            num_subjects = 9
            sessions = ['T', 'E']
            save_path = folder_name + '/' + dataset + '/raw'
            if save_path is not None:
                if not os.path.exists(save_path):
                    os.makedirs(save_path)

            for session in sessions:
                for person in range(1, num_subjects+1):
                    file_name = '/A{:02d}{}.mat'.format(person, session)
                    if os.path.exists(save_path+file_name):
                        os.remove(save_path+file_name) # if exist, remove file
                    print('\n===Download is being processed on session: {} subject: {}==='.format(session, person))
                    url = 'https://lampx.tugraz.at/~bci/database/001-2014'+file_name
                    print('save to: '+save_path+file_name)
                    wget.download(url, save_path+file_name)
            print('\nDone!')
        except:
            raise Exception('Path Error: file does not exist, please direccly download at http://bnci-horizon-2020.eu/database/data-sets')
    elif dataset == 'SMR_BCI':
        try:
            num_subjects = 14
            sessions = ['T', 'E']
            save_path = folder_name + '/' + dataset + '/raw'
            if save_path is not None:
                if not os.path.exists(save_path):
                    os.makedirs(save_path)
            for session in sessions:
                for person in range(1, num_subjects+1):
                    file_name = '/S{:02d}{}.mat'.format(person, session)
                    if os.path.exists(save_path+file_name):
                        os.remove(save_path+file_name) # if exist, remove file
                    print('\n===Download is being processed on session: {} subject: {}==='.format(session, person))
                    url = 'https://lampx.tugraz.at/~bci/database/002-2014'+file_name
                    print('save to: '+save_path+file_name)
                    wget.download(url,  save_path+file_name)
            print('\nDone!')
        except:
            raise Exception('Path Error: file does not exist, please direccly download at http://bnci-horizon-2020.eu/database/data-sets')

class DataLoader:
    def __init__(self, dataset, train_type=None, data_type=None, num_class=2, subject=None, data_format=None, dataset_path='/datasets', **kwargs):

        self.dataset = dataset #Dataset name: 'OpenBMI', 'SMR_BCI', 'BCIC2a'
        self.train_type = train_type # 'subject_dependent', 'subject_independent'
        self.data_type = data_type # 'fbcsp', 'spectral_spatial', 'time_domain'
        self.dataset_path = dataset_path
        self.subject = subject # id, start at 1
        self.data_format = data_format # 'channels_first', 'channels_last'
        self.fold = None # fold, start at 1
        self.prefix_name = 'S'
        self.num_class = num_class
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])


        self.path = self.dataset_path+'/'+self.dataset+'/'+self.data_type+'/'+str(self.num_class)+'_class/'+self.train_type
    
    def _change_data_format(self, X):
        if self.data_format == 'NCTD':
            # (#n_trial, #channels, #time, #depth)
            X = X.reshape(X.shape[0], X.shape[1], X.shape[2], 1)
        elif self.data_format == 'NDCT':
            # (#n_trial, #depth, #channels, #time)
            X = X.reshape(X.shape[0], 1, X.shape[1], X.shape[2])
        elif self.data_format == 'NTCD':
            # (#n_trial, #time, #channels, #depth)
            X = X.reshape(X.shape[0], X.shape[1], X.shape[2], 1)
            X = np.swapaxes(X, 1, 3)
        elif self.data_format == 'NSHWD':
            # (#n_trial, #Freqs, #height, #width, #depth)
            X = zero_padding(X)
            X = X.reshape(X.shape[0], X.shape[1], X.shape[2], X.shape[3], 1)
        elif self.data_format == None:
            pass
        else:
            raise Exception('Value Error: data_format requires None, \'NCTD\', \'NDCT\', \'NTCD\' or \'NSHWD\', found data_format={}'.format(self.data_format))
        print('change data_format to \'{}\', new dimention is {}'.format(self.data_format, X.shape))
        return X

    def load_train_set(self, fold, **kwargs):
        self.fold = fold
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
    
        # load 
        X, y =  np.array([]),  np.array([])
        try:
            self.file_x = self.path+'/X_train_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            self.file_y = self.path+'/y_train_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            X = self._change_data_format(np.load(self.file_x))
            y = np.load(self.file_y)
        except:
            raise Exception('Path Error: file does not exist, please check this path {}, and {}'.format(self.file_x, self.file_y))
        return X, y

    def load_val_set(self, fold, **kwargs):
        self.fold = fold
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])

        # load 
        X, y =  np.array([]),  np.array([])
        try:
            self.file_x = self.path+'/X_val_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            self.file_y = self.path+'/y_val_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            X = self._change_data_format(np.load(self.file_x))
            y = np.load(self.file_y)
        except:
            raise Exception('Path Error: file does not exist, please check this path {}, and {}'.format(self.file_x, self.file_y))
        return X, y
    
    def load_test_set(self, fold, **kwargs):
        self.fold = fold
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])

        # load 
        X, y =  np.array([]),  np.array([])
        try:
            self.file_x = self.path+'/X_test_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            self.file_y = self.path+'/y_test_{}{:03d}_fold{:03d}.npy'.format(self.prefix_name, self.subject, self.fold)
            X = self._change_data_format(np.load(self.file_x))
            y = np.load(self.file_y)
        except:
            raise Exception('Path Error: file does not exist, please check this path {}, and {}'.format(self.file_x, self.file_y))
        return X, y

def compute_class_weight(y_train):
    """compute class balancing

    Args:
        y_train (list, ndarray): [description]

    Returns:
        (dict): class weight balancing
    """
    return dict(zip(np.unique(y_train), 
                    class_weight.compute_class_weight(class_weight='balanced',
                                                      classes=np.unique(y_train),
                                                      y=y_train))) 
        
def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')

class TimeHistory(tf.keras.callbacks.Callback):
    def __init__(self, save_path=None):
        self.save_path = save_path
    def on_train_begin(self, logs={}):
        self.logs = []
        if self.save_path:
            write_log(filepath=self.save_path, data=['time_log'], mode='w')
    def on_epoch_begin(self, epoch, logs={}):
        self.start_time = time.time()
    def on_epoch_end(self, epoch, logs={}):
        time_diff = time.time()-self.start_time
        self.logs.append(time_diff)
        if self.save_path:
            write_log(filepath=self.save_path, data=[time_diff], mode='a')

def write_log(filepath='test.log', data=[], mode='w'):
    '''
    filepath: path to save
    data: list of data
    mode: a = update data to file, w = write a new file
    '''
    try:
        with open(filepath, mode) as csvFile:
            writer = csv.writer(csvFile)
            writer.writerow(data)
    except IOError:
        raise Exception('I/O error')

def zero_padding(data, pad_size=4):
    if len(data.shape) != 4:
        raise Exception('Dimension is not match!, must have 4 dims')
    new_shape = int(data.shape[2]+(2*pad_size))
    data_pad = np.zeros((data.shape[0], data.shape[1], new_shape, new_shape))
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            data_pad[i,j,:,:] = np.pad(data[i,j,:,:], [pad_size, pad_size], mode='constant')
    print(data_pad.shape)
    return data_pad 


def butter_bandpass_filter(data, lowcut, highcut, fs, order):
    nyq = 0.5 * fs
    low = lowcut / nyq
    high = highcut / nyq
    b, a = butter(order, [low, high], btype='band')
    y = filtfilt(b, a, data)
    return y

def resampling(data, new_smp_freq, data_len):
    if len(data.shape) != 3:
        raise Exception('Dimesion error', "--> please use three-dimensional input")
    new_smp_point = int(data_len*new_smp_freq)
    data_resampled = np.zeros((data.shape[0], data.shape[1], new_smp_point))
    for i in range(data.shape[0]):
        for j in range(data.shape[1]):
            data_resampled[i,j,:] = signal.resample(data[i,j,:], new_smp_point)
    return data_resampled

def psd_welch(data, smp_freq):
    if len(data.shape) != 3:
        raise Exception("Dimension Error, must have 3 dimension")
    n_samples,n_chs,n_points = data.shape
    data_psd = np.zeros((n_samples,n_chs,89))
    for i in range(n_samples):
        for j in range(n_chs):
            freq, power_den = signal.welch(data[i,j], smp_freq, nperseg=n_points)
            index = np.where((freq>=8) & (freq<=30))[0].tolist()
            # print("the length of---", len(index))
            data_psd[i,j] = power_den[index]
    return data_psd


import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Conv2D, Dense, Input, Flatten, Dropout, Concatenate
from tensorflow.keras.callbacks import CSVLogger, ModelCheckpoint, ReduceLROnPlateau, EarlyStopping

import numpy as np
import os
import time
from sklearn.metrics import classification_report, f1_score
from min2net.utils import TimeHistory, compute_class_weight

class SpectralSpatialCNN:
    def __init__(self,
                input_shape=(28,28,1),
                num_class=2,
                loss='sparse_categorical_crossentropy',
                epochs=200,
                batch_size=100,
                optimizer = Adam(beta_1=0.9, beta_2=0.999, epsilon=1e-08),
                lr=1e-5,
                min_lr=1e-5,
                factor=0.25,
                patience=10,
                es_patience=20,
                verbose=1,
                log_path='logs',
                model_name='SpectralSpatialCNN',
                **kwargs):
        self.input_shape = input_shape
        self.num_class = num_class
        self.loss = loss
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.optimizer.lr = lr
        self.lr = lr
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.es_patience = es_patience
        self.verbose = verbose
        self.log_path = log_path
        self.model_name = model_name
        self.weights_dir = log_path+'/'+model_name+'_out_weights.h5'
        self.csv_dir = log_path+'/'+model_name+'_out_log.log'
        self.time_log = log_path+'/'+model_name+'_time_log.csv'

            
        # use **kwargs to set the new value of below args.
        self.n_subbands = 20
        self.dropout_rate = 0.5
        self.f1_average = 'binary' if self.num_class == 2 else 'macro'
        self.data_format = 'channels_last'
        self.shuffle = False
        self.metrics = 'accuracy'
        self.monitor = 'val_loss'
        self.mode = 'min'
        self.save_best_only = True
        self.save_weight_only = True
        self.seed = 1234
        self.class_balancing = False
        self.class_weight = None
        
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
        
        np.random.seed(self.seed)
        tf.random.set_seed(self.seed)
        K.set_image_data_format(self.data_format)
        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)

            
    def __cnn_model(self, model_input): 
        # DeepConvNet
        model = Conv2D(filters=10, kernel_size=(3, 3), activation='relu', padding="same")(model_input) 
        model = Conv2D(filters=14, kernel_size=(3, 3), activation='relu', padding="same")(model)
        model = Conv2D(filters=18, kernel_size=(3, 3), activation='relu', padding="same")(model)
        model = Flatten()(model)
        model = Dense(256)(model)
        return model

    def build(self): 
        input_data    = [Input(shape=self.input_shape) for i in range(self.n_subbands)]
        model_sub     = [self.__cnn_model(input_data[i]) for i in range(self.n_subbands)] 
        out_sub_model = Concatenate()([model_sub[i] for i in range(self.n_subbands)])
        sub_model     = Dense(1024)(out_sub_model)
        sub_model     = Dropout(self.dropout_rate)(sub_model)
        sub_model     = Dense(self.num_class, activation='softmax')(sub_model)
        model         = Model(inputs=input_data, outputs=sub_model)
        return model

    def fit(self, X_train, y_train, X_val, y_val):
        if X_train.ndim != 5:
            raise Exception('ValueError: `X_train` is incompatible: expected ndim=5, found ndim='+str(X_train.ndim))
        elif X_val.ndim != 5:
            raise Exception('ValueError: `X_val` is incompatible: expected ndim=5, found ndim='+str(X_val.ndim))
        
        self.input_shape = (X_train.shape[2], X_train.shape[3], X_train.shape[4])
        self.n_subbands = X_train.shape[1]
        
        csv_logger = CSVLogger(self.csv_dir)
        time_callback = TimeHistory(self.time_log)
        checkpointer = ModelCheckpoint(monitor=self.monitor, filepath=self.weights_dir, verbose=self.verbose, 
                                       save_best_only=self.save_best_only, save_weight_only=self.save_weight_only)
        reduce_lr = ReduceLROnPlateau(monitor=self.monitor, patience=self.patience, factor=self.factor, mode=self.mode, 
                                      verbose=self.verbose, min_lr=self.min_lr)
        es = EarlyStopping(monitor=self.monitor, mode=self.mode, verbose=self.verbose, patience=self.es_patience)
        model = self.build()
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
        model.summary()
        
        if self.class_balancing: # compute_class_weight if class_balancing is True
            self.class_weight = compute_class_weight(y_train)
        else:
            self.class_weight = None
            
        model.fit([X_train[:,i,:,:,:] for i in range(self.n_subbands)], y_train,
                batch_size=self.batch_size,  shuffle=self.shuffle, class_weight=self.class_weight,
                epochs=self.epochs, validation_data=([X_val[:,i,:,:,:] for i in range(self.n_subbands)], y_val),
                callbacks=[checkpointer,csv_logger,reduce_lr,es,time_callback])

    def predict(self, X_test, y_test):
        if X_test.ndim != 5:
            raise Exception('ValueError: `X_test` is incompatible: expected ndim=5, found ndim='+str(X_test.ndim))
        model = self.build()
        model.load_weights(self.weights_dir)
        self.n_subbands = X_test.shape[1]
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)

        start = time.time()
        y_pred = model.predict([X_test[:,i,:,:,:] for i in range(self.n_subbands)])
        end = time.time()
        loss, acc = model.evaluate(x=[X_test[:,i,:,:,:] for i in range(self.n_subbands)], y=y_test, 
                                   batch_size=self.batch_size, verbose=self.verbose)
        y_pred_argm = np.argmax(y_pred, axis=1)
        print("F1-score is comptured basen on {}".format(self.f1_average))
        f1 = f1_score(y_test, y_pred_argm, average=self.f1_average)
        print(classification_report(y_test, y_pred_argm))
        evaluation = {'loss': loss, 
                      'accuracy': acc, 
                      'f1-score': f1, 
                      'prediction_time': end-start}
        Y = {'y_true': y_test, 
             'y_pred': y_pred_argm}
        return Y, evaluation


import numpy as np
import scipy.io as sio
from min2net.utils import resampling
from min2net.preprocessing.config import CONSTANT
# load variable form config file
CONSTANT = CONSTANT['SMR_BCI']
orig_chs = CONSTANT['orig_chs']
orig_smp_freq = CONSTANT['orig_smp_freq']
trial_len = CONSTANT['trial_len'] 
n_chs = CONSTANT['n_chs'] # 15
window_len = CONSTANT['trial_len']*CONSTANT['orig_smp_freq'] # 8*512

def read_raw(PATH, subject , training, id_chosen_chs):
    if training:
        mat = sio.loadmat(PATH+'/S'+str(subject).zfill(2)+'T.mat')['data']
        n_trials = CONSTANT['n_trials_tr'] # 100
        label = np.zeros(n_trials)
        data = np.zeros((n_trials, n_chs, window_len))
    else:
        mat = sio.loadmat(PATH+'/S'+str(subject).zfill(2)+'E.mat')['data']
        n_trials = CONSTANT['n_trials_te'] # 60
        label = np.zeros(n_trials)
        data = np.zeros((n_trials, n_chs, window_len))
    NO_valid_trial = 0
    for ii in range(0,mat.size):
        mat_1 = mat[0,ii]
        mat_2 = [mat_1[0,0]]
        mat_info = mat_2[0]
        _X = mat_info[0]
        _trial = mat_info[1]
        _y = mat_info[2]
        _fs = mat_info[3]
        _classes = mat_info[4]
        for trial in range(0, _trial.size):
            # class 1 (right hand) and class 2 (feet) 
            _data = np.transpose(_X[_trial[0][trial]:int(_trial[0][trial]+window_len),id_chosen_chs])
            _label = int(_y[0][trial])
            data[NO_valid_trial,:,:] =  _data
            label[NO_valid_trial] = _label
            NO_valid_trial +=1
    return data, label-1

def chanel_selection(sel_chs): 
    chs_id = []
    for name_ch in sel_chs:
        ch_id = np.where(np.array(orig_chs) == name_ch)[0][0]
        chs_id.append(ch_id)
        print('chosen_channel:', name_ch, '---', 'Index_is:', ch_id)
    return chs_id

def load_crop_data(PATH, subject, start, stop, new_smp_freq, id_chosen_chs):
    start_time = int(start*new_smp_freq) # 4*
    stop_time = int(stop*new_smp_freq) # 8*
    X_train, y_tr = read_raw(PATH=PATH, subject=subject, training=True, id_chosen_chs=id_chosen_chs)
    X_test, y_te = read_raw(PATH=PATH, subject=subject, training=False, id_chosen_chs=id_chosen_chs)
    if new_smp_freq < orig_smp_freq:
        X_train = resampling(X_train, new_smp_freq, trial_len)
        X_test = resampling(X_test, new_smp_freq, trial_len)
    X_train = X_train[:,:,start_time:stop_time]
    X_test = X_test[:,:,start_time:stop_time]
    return X_train, y_tr, X_test, y_te  


from .MIN2Net import MIN2Net
from .MIN2Net_without_triplet import MIN2Net_without_triplet
from .DeepConvNet import DeepConvNet
from .EEGNet import EEGNet
from .SVM import SVM
from .SpectralSpatialCNN import SpectralSpatialCNN
from .MIN2Net_without_decoder import MIN2Net_without_decoder


import tensorflow as tf
from tensorflow.keras import backend as K
from tensorflow.keras.optimizers import Adam
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Conv2D, Dense, Flatten, Dropout, BatchNormalization, Activation, MaxPooling2D, DepthwiseConv2D, AveragePooling2D, SeparableConv2D
from tensorflow.keras.constraints import max_norm
from tensorflow.keras.callbacks import CSVLogger, ModelCheckpoint, ReduceLROnPlateau, EarlyStopping

import numpy as np
import os
import time
from sklearn.metrics import classification_report, f1_score
from min2net.utils import TimeHistory, compute_class_weight

class EEGNet:
    def __init__(self,
                input_shape=(1,20,400),
                num_class=2,
                loss='sparse_categorical_crossentropy',
                epochs=200,
                batch_size=100,
                optimizer = Adam(beta_1=0.9, beta_2=0.999, epsilon=1e-08),
                lr=0.01,
                min_lr=0.01,
                factor=0.25,
                patience=10,
                es_patience=20,
                verbose=1,
                log_path='logs',
                model_name='EEGNet',
                **kwargs):
        self.input_shape = input_shape
        self.num_class = num_class
        self.loss = loss
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.optimizer.lr = lr
        self.lr = lr
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.es_patience = es_patience
        self.verbose = verbose
        self.log_path = log_path
        self.model_name = model_name
        self.weights_dir = log_path+'/'+model_name+'_out_weights.h5'
        self.csv_dir = log_path+'/'+model_name+'_out_log.log'
        self.time_log = log_path+'/'+model_name+'_time_log.csv'

        # use **kwargs to set the new value of below args.
        self.kernLength = 200
        self.F1 = 8
        self.D = 2
        self.F2 = int(self.F1*self.D)
        self.norm_rate = 0.25
        self.dropout_rate = 0.5
        self.f1_average = 'binary' if self.num_class == 2 else 'macro'
        self.data_format = 'channels_first'
        self.shuffle = False
        self.metrics = 'accuracy'
        self.monitor = 'val_loss'
        self.mode = 'min'
        self.save_best_only = True
        self.save_weight_only = True
        self.seed = 1234
        self.class_balancing = False
        self.class_weight = None

        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
            
        if self.data_format == 'channels_first':
            self.Chans = self.input_shape[1]
            self.Samples = self.input_shape[2]
        else:
            self.Chans = self.input_shape[0]
            self.Samples = self.input_shape[1]

        np.random.seed(self.seed)
        tf.random.set_seed(self.seed)
        K.set_image_data_format(self.data_format)
        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)

    def build(self):
        input1       = Input(shape=self.input_shape)

        ##################################################################
        block1       = Conv2D(self.F1, (1, self.kernLength), padding='same',
                              input_shape=self.input_shape,
                              use_bias=False)(input1)
        block1       = BatchNormalization()(block1)
        block1       = DepthwiseConv2D((self.Chans, 1), use_bias=False,
                                       depth_multiplier=self.D,
                                       depthwise_constraint=max_norm(1.))(block1)
        block1       = BatchNormalization()(block1)
        block1       = Activation('elu')(block1)
        block1       = AveragePooling2D((1, 4))(block1)
        block1       = Dropout(self.dropout_rate)(block1)

        block2       = SeparableConv2D(self.F2, (1, self.kernLength//4),
                                       use_bias=False, padding='same')(block1)
        block2       = BatchNormalization()(block2)
        block2       = Activation('elu')(block2)
        block2       = AveragePooling2D((1, 8))(block2)
        block2       = Dropout(self.dropout_rate)(block2)

        flatten      = Flatten(name='flatten')(block2)

        dense        = Dense(self.num_class, name='dense',
                             kernel_constraint=max_norm(self.norm_rate))(flatten)
        softmax      = Activation('softmax', name='softmax')(dense)

        return Model(inputs=input1, outputs=softmax)

    def fit(self, X_train, y_train, X_val, y_val):

        if X_train.ndim != 4:
            raise Exception('ValueError: `X_train` is incompatible: expected ndim=4, found ndim='+str(X_train.ndim))
        elif X_val.ndim != 4:
            raise Exception('ValueError: `X_val` is incompatible: expected ndim=4, found ndim='+str(X_val.ndim))

        self.input_shape = X_train.shape[1:]
        if self.data_format == 'channels_first':
            self.Chans = self.input_shape[1]
            self.Samples = self.input_shape[2]
        else:
            self.Chans = self.input_shape[0]
            self.Samples = self.input_shape[1]
        
        csv_logger = CSVLogger(self.csv_dir)
        time_callback = TimeHistory(self.time_log)
        checkpointer = ModelCheckpoint(monitor=self.monitor, filepath=self.weights_dir, verbose=self.verbose, 
                                       save_best_only=self.save_best_only, save_weight_only=self.save_weight_only)
        reduce_lr = ReduceLROnPlateau(monitor=self.monitor, patience=self.patience, factor=self.factor, 
                                      mode=self.mode, verbose=self.verbose, min_lr=self.min_lr)
        es = EarlyStopping(monitor=self.monitor, mode=self.mode, verbose=self.verbose, patience=self.es_patience)

        model = self.build()
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)
        model.summary()
        print("The first kernel size is (1, {})".format(self.kernLength))
        
        if self.class_balancing: # compute_class_weight if class_balancing is True
            self.class_weight = compute_class_weight(y_train)
        else:
            self.class_weight = None         
            
        model.fit(X_train, y_train,
                  batch_size=self.batch_size, shuffle=self.shuffle,
                  epochs=self.epochs, validation_data=(X_val, y_val), class_weight=self.class_weight,
                  callbacks=[checkpointer,csv_logger,reduce_lr,es, time_callback])

    def predict(self, X_test, y_test):

        if X_test.ndim != 4:
            raise Exception('ValueError: `X_test` is incompatible: expected ndim=4, found ndim='+str(X_test.ndim))

        model = self.build()
        model.load_weights(self.weights_dir)
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics)

        start = time.time()
        y_pred = model.predict(X_test)
        end = time.time()
        loss, accuracy = model.evaluate(x=X_test, y=y_test, batch_size=self.batch_size, verbose=self.verbose)
        y_pred_argm = np.argmax(y_pred, axis=1)
        print(classification_report(y_test, y_pred_argm))
        print("F1-score is computed based on {}".format(self.f1_average))
        f1 = f1_score(y_test, y_pred_argm, average=self.f1_average)
        evaluation = {'loss': loss, 
                      'accuracy': accuracy, 
                      'f1-score': f1, 
                      'prediction_time': end-start}
        Y = {'y_true': y_test, 
             'y_pred': y_pred_argm}
        return Y, evaluation


import numpy as np
from sklearn.model_selection import StratifiedKFold
import os
from min2net.utils import butter_bandpass_filter
from min2net.preprocessing.OpenBMI import raw
from min2net.preprocessing.config import CONSTANT
CONSTANT = CONSTANT['OpenBMI']
raw_path = CONSTANT['raw_path']
n_subjs = CONSTANT['n_subjs']
orig_smp_freq = CONSTANT['orig_smp_freq']
MI_len = CONSTANT['MI']['len']

def subject_dependent_setting(k_folds, pick_smp_freq, bands, order, save_path, num_class=2, sel_chs=None):
    sel_chs = CONSTANT['sel_chs'] if sel_chs == None else sel_chs
    n_folds = k_folds
    save_path = save_path + '/OpenBMI/time_domain/{}_class/subject_dependent'.format(num_class)

    id_chosen_chs = raw.chanel_selection(sel_chs)
    X_train_all, y_train_all, X_test_all, y_test_all = __load_OpenBMI(raw_path, n_subjs, pick_smp_freq, num_class, MI_len, id_chosen_chs)
    for directory in [save_path]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    # Carry out subject-dependent setting with k-fold cross validation
    for person, (X_tr, y_tr, X_te, y_te) in enumerate(zip(X_train_all, y_train_all, X_test_all, y_test_all)):
        if len(X_tr.shape) != 3:
            raise Exception('Dimension Error, must have 3 dimension')

        skf = StratifiedKFold(n_splits=n_folds, random_state=42, shuffle=True)
        for fold, (train_index, val_index) in enumerate(skf.split(X_tr , y_tr)):
            print('FOLD:', fold+1, 'TRAIN:', len(train_index), 'VALIDATION:', len(val_index))
            X_tr_cv, X_val_cv = X_tr[train_index], X_tr[val_index]
            y_tr_cv, y_val_cv = y_tr[train_index], y_tr[val_index]

            print('Band-pass filtering from {} to {} Hz.'.format(bands[0],  bands[1]))
            X_tr_fil = butter_bandpass_filter(X_tr_cv,  bands[0],  bands[1], pick_smp_freq, order)
            X_val_fil = butter_bandpass_filter(X_val_cv,  bands[0],  bands[1], pick_smp_freq, order)
            X_te_fil = butter_bandpass_filter(X_te,  bands[0],  bands[1], pick_smp_freq, order)
            print('Check dimension of training data {}, val data {} and testing data {}'.format(X_tr_fil.shape, X_val_fil.shape, X_te_fil.shape))
            SAVE_NAME = 'S{:03d}_fold{:03d}'.format(person+1, fold+1)
            __save_data_with_valset(save_path, SAVE_NAME, X_tr_fil, y_tr_cv, X_val_fil, y_val_cv, X_te_fil, y_te)
            print('The preprocessing of subject {} from fold {} is DONE!!!'.format(person+1, fold+1))
                      
def subject_independent_setting(k_folds, pick_smp_freq, bands, order, save_path, num_class=2, sel_chs=None):
    sel_chs = CONSTANT['sel_chs'] if sel_chs == None else sel_chs
    n_folds = k_folds
    save_path = save_path + '/OpenBMI/time_domain/{}_class/subject_independent'.format(num_class)

    id_chosen_chs = raw.chanel_selection(sel_chs)
    X_train_all, y_train_all, X_test_all, y_test_all = __load_OpenBMI(raw_path, n_subjs, pick_smp_freq, num_class, MI_len, id_chosen_chs)
    for directory in [save_path]:
        if not os.path.exists(directory):
            os.makedirs(directory)

    # Carry out subject-independent setting with 5-fold cross validation
    for person, (X_val, y_val, X_te, y_te) in enumerate(zip(X_train_all, y_train_all, X_test_all, y_test_all)):
        train_subj = [i for i in range(n_subjs)]
        train_subj = np.delete(train_subj, person) # remove test subject

         # Generating fake data to used for k-fold cross-validation only
        fake_tr = np.zeros((len(train_subj), 2))
        fake_tr_la = np.zeros((len(train_subj)))

        skf = StratifiedKFold(n_splits=n_folds, random_state=42, shuffle=True)
        for fold, (train_ind, val_ind) in enumerate(skf.split(fake_tr , fake_tr_la)):
            print('FOLD:', fold+1, 'TRAIN:', len(train_ind), 'VALIDATION:', len(val_ind))
            train_index, val_index = train_subj[train_ind], train_subj[val_ind]
            X_train_cat = np.concatenate((X_train_all[train_index], X_test_all[train_index]), axis=0)
            X_val_cat = np.concatenate((X_train_all[val_index], X_test_all[val_index]), axis=0)
            y_train_cat = np.concatenate((y_train_all[train_index], y_test_all[train_index]), axis=0)
            y_val_cat = np.concatenate((y_train_all[val_index], y_test_all[val_index]), axis=0)

            X_train = X_train_cat.reshape(-1, X_train_cat.shape[2], X_train_cat.shape[3])
            y_train = y_train_cat.reshape(-1)
            X_val = X_val_cat.reshape(-1, X_val_cat.shape[2], X_val_cat.shape[3])
            y_val = y_val_cat.reshape(-1)
            X_test = X_te
            y_test = y_te

            # Performing bandpass-filtering
            print('Band-pass filtering from {} to {} Hz.'.format(bands[0],  bands[1]))
            X_train_fil =  butter_bandpass_filter(X_train, bands[0],  bands[1], pick_smp_freq, order)
            X_val_fil = butter_bandpass_filter(X_val, bands[0],  bands[1], pick_smp_freq, order)
            X_test_fil = butter_bandpass_filter(X_test, bands[0],  bands[1], pick_smp_freq, order)

            print('Verify the final dimesion of training data {}, val data {} and testing data {}'.format(X_train_fil.shape, X_val_fil.shape,X_test_fil.shape))
            print("Check class number Tr {} Val {} and Te {}".format(np.unique(y_train),np.unique(y_val),np.unique(y_test)))
            SAVE_NAME = 'S{:03d}_fold{:03d}'.format(person+1, fold+1)
            __save_data_with_valset(save_path, SAVE_NAME, X_train_fil, y_train, X_val_fil, y_val, X_test_fil, y_test)
            print('The preprocessing of subject {} from fold {} is DONE!!!'.format(person+1, fold+1))

def __load_OpenBMI(PATH, n_subjs, new_smp_freq, num_class, MI_len, id_chosen_chs):
    X_train_all, y_train_all, X_test_all, y_test_all = raw.load_crop_data(PATH, n_subjs, new_smp_freq, num_class, MI_len, id_chosen_chs)
    return X_train_all, y_train_all, X_test_all, y_test_all

def __save_data_with_valset(save_path, NAME, X_train, y_train, X_val, y_val, X_test, y_test):
    np.save(save_path+'/X_train_'+NAME+'.npy', X_train)
    np.save(save_path+'/X_val_'+NAME+'.npy', X_val)
    np.save(save_path+'/X_test_'+NAME+'.npy', X_test)
    np.save(save_path+'/y_train_'+NAME+'.npy', y_train)
    np.save(save_path+'/y_val_'+NAME+'.npy', y_val)
    np.save(save_path+'/y_test_'+NAME+'.npy', y_test)
    print('save DONE')



from sklearn.metrics import classification_report, f1_score
from sklearn.svm import SVC
from sklearn.model_selection import GridSearchCV, PredefinedSplit
from joblib import dump, load
import numpy as np
import os
import time
from min2net.utils import write_log

class SVM:
    def __init__(self,
                log_path='logs',
                model_name='SVM',
                tuned_parameters=None,
                num_class=2,
                **kwargs):
        if tuned_parameters == None:
            self.tuned_parameters = [{'kernel': ['rbf'],
                                      'gamma': [1e-2, 1e-3],
                                      'C': [0.001, 0.01, 0.1, 1, 10, 100, 1000]
                                     },
                                    {'kernel': ['sigmoid'],
                                     'gamma': [1e-2, 1e-3],
                                     'C': [0.001, 0.01, 0.1, 1, 10, 100, 1000]
                                    },
                                    {'kernel': ['linear'],
                                     'gamma': [1e-2, 1e-3],
                                     'C':[0.001, 0.01, 0.1, 1, 10, 100, 1000]
                                    }]
        else:
            self.tuned_parameters = tuned_parameters
        self.seed = 1234
        self.num_class = num_class
        self.f1_average = 'binary' if self.num_class == 2 else 'macro'

        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
            
        np.random.seed(self.seed)

        if not os.path.exists(log_path):
            os.makedirs(log_path)
        self.model_path = log_path+'/'+model_name+'_model.gz'
        self.time_log = log_path+'/'+model_name+'_time_log.csv'

    def fit(self, X_train, y_train, X_val, y_val):

        if X_train.ndim != 2:
            raise Exception('ValueError: `X_train` is incompatible: expected ndim=4, found ndim='+str(X_train.ndim))
        elif X_val.ndim != 2:
            raise Exception('ValueError: `X_val` is incompatible: expected ndim=4, found ndim='+str(X_val.ndim))

        print('Dimension of training set is: {} and label is: {}'.format(X_train.shape, y_train.shape))
        print('Dimension of validation set is: {} and label is: {}'.format(X_val.shape, y_val.shape))

        X_all = np.concatenate((X_train, X_val),axis=0)
        y_all = np.concatenate((y_train, y_val),axis=0)

        # Create a list where train data indices are -1 and validation data indices are 0
        tr_index = np.full((X_train.shape[0]), -1)
        val_index = np.full((X_val.shape[0]), 0)
        split_index = np.concatenate((tr_index, val_index), axis=0).tolist()
        # Use the list to create PredefinedSplit
        pds = PredefinedSplit(test_fold = split_index)
        clf = GridSearchCV(estimator=SVC(), param_grid=self.tuned_parameters, cv=pds, scoring = 'accuracy')
        start = time.time()
        clf.fit(X_all , y_all)
        end = time.time()
        #Clasifying with an optimal parameter set
        Optimal_params = clf.best_params_
        print(Optimal_params)
        classifier = SVC(**Optimal_params)
        classifier.fit(X_train, y_train)
        dump(classifier, self.model_path)
        write_log(filepath=self.time_log, data=['time_log'], mode='w')
        write_log(filepath=self.time_log, data=[end-start], mode='a')

    def predict(self, X_test, y_test):

        if X_test.ndim != 2:
            raise Exception('ValueError: `X_test` is incompatible: expected ndim=4, found ndim='+str(X_test.ndim))

        print('Dimesion of testing set is: {} and label is: {}'.format(X_test.shape, y_test.shape))
        classifier = load(self.model_path)
        svm_acc = classifier.score(X_test, y_test)
        print('===Test Score for Optimized Parameters===:', svm_acc)
        start = time.time()
        y_true, y_pred = y_test, classifier.predict(X_test)
        end = time.time()
        print(classification_report(y_true, y_pred))
        n_classes = len(np.unique(y_test))
        print("F1-score is computed based on {}".format(self.f1_average))
        f1 = f1_score(y_true, y_pred, average=self.f1_average)
        evaluation = {'accuracy': svm_acc, 
                      'f1-score': f1, 
                      'prediction_time': end-start}
        Y = {'y_true': y_true, 
             'y_pred': y_pred}
        return Y, evaluation



from tensorflow.keras.layers import Concatenate, AveragePooling2D, BatchNormalization, Conv2D, Conv2DTranspose, Dense, Input, Reshape, Flatten
from tensorflow.keras import models
from tensorflow.keras.constraints import max_norm
from tensorflow.keras.callbacks import CSVLogger, ModelCheckpoint, ReduceLROnPlateau
from tensorflow.keras.callbacks import EarlyStopping
from tensorflow.keras.optimizers import Adam
import tensorflow as tf
from tensorflow.keras import backend as K

import numpy as np
import time
import os
from sklearn.metrics import classification_report, f1_score
from min2net.loss import triplet_loss, SparseCategoricalCrossentropy
from min2net.utils import TimeHistory, compute_class_weight

class MIN2Net_without_decoder:
    def __init__(self,
                input_shape=(1,400,20), 
                num_class=2, 
                loss=[triplet_loss(margin=1.0), 'sparse_categorical_crossentropy'],
                loss_weights=[1., 1.], 
                latent_dim = None,
                epochs=200,
                batch_size=100,
                optimizer = Adam(beta_1=0.9, beta_2=0.999, epsilon=1e-08),
                lr=1e-2,
                min_lr=1e-3,
                factor=0.25,
                patience=20, 
                es_patience=50,
                verbose=1,
                log_path='logs',
                model_name='MIN2Net_without_decoder', 
                **kwargs):
        D, T, C = input_shape
        self.latent_dim = latent_dim if latent_dim is not None else C if num_class==2 else 64
        self.num_class = num_class
        self.input_shape = input_shape
        self.loss = loss
        self.loss_weights = loss_weights
        self.epochs = epochs
        self.batch_size = batch_size
        self.optimizer = optimizer
        self.optimizer.lr = lr
        self.lr = lr
        self.min_lr = min_lr
        self.factor = factor
        self.patience = patience
        self.es_patience = es_patience
        self.verbose = verbose
        self.log_path = log_path
        self.model_name = model_name
        self.weights_dir = log_path+'/'+model_name+'_out_weights.h5'
        self.csv_dir = log_path+'/'+model_name+'_out_log.log'
        self.time_log = log_path+'/'+model_name+'_time_log.csv'

        # use **kwargs to set the new value of below args.
        self.f1_average = 'binary' if self.num_class == 2 else 'macro'
        self.data_format = 'channels_last'
        self.shuffle = False
        self.metrics = 'accuracy'
        self.monitor = 'val_loss'
        self.mode = 'min'
        self.save_best_only = True
        self.save_weight_only = True
        self.seed = 1234
        self.class_balancing = False
        # 'set params'
        self.subsampling_size = 100
        self.subsampling_size = 100
        self.pool_size_1 = (1,T//self.subsampling_size)
        self.pool_size_2 = (1,4)
        self.filter_1 = C
        self.filter_2 = 10
        
        for k in kwargs.keys():
            self.__setattr__(k, kwargs[k])
        
        self.flatten_size = T//self.pool_size_1[1]//self.pool_size_2[1]
        
        np.random.seed(self.seed)
        tf.random.set_seed(self.seed)
        K.set_image_data_format(self.data_format)
        if not os.path.exists(self.log_path):
            os.makedirs(self.log_path)

    def build(self):
        'encoder --> DCNN'
        encoder_input  = Input(self.input_shape)
        en_conv        = Conv2D(self.filter_1, (1, 64), activation='elu', padding="same", 
                                kernel_constraint=max_norm(2., axis=(0, 1, 2)))(encoder_input)
        en_conv        = BatchNormalization(axis=3, epsilon=1e-05, momentum=0.1)(en_conv)
        en_conv        = AveragePooling2D(pool_size=self.pool_size_1)(en_conv)  
        en_conv        = Conv2D(self.filter_2, (1, 32), activation='elu', padding="same", 
                                kernel_constraint=max_norm(2., axis=(0, 1, 2)))(en_conv)
        en_conv        = BatchNormalization(axis=3, epsilon=1e-05, momentum=0.1)(en_conv)
        en_conv        = AveragePooling2D(pool_size=self.pool_size_2)(en_conv)
        en_conv        = Flatten()(en_conv)
        encoder_output = Dense(self.latent_dim, kernel_constraint=max_norm(0.5))(en_conv)
   

        z              = Dense(self.num_class, activation='softmax', kernel_constraint=max_norm(0.5), 
                               name='classifier')(encoder_output)

        return models.Model(inputs=encoder_input, outputs=[encoder_output, z], 
                            name='MIN2Net_without_decoder')

    
    def fit(self, X_train, y_train, X_val, y_val):

        if X_train.ndim != 4:
            raise Exception('ValueError: `X_train` is incompatible: expected ndim=4, found ndim={}'.format(X_train.ndim))
        elif X_val.ndim != 4:
            raise Exception('ValueError: `X_val` is incompatible: expected ndim=4, found ndim={}'.format(X_val.ndim))

        csv_logger = CSVLogger(self.csv_dir)
        time_callback = TimeHistory(self.time_log)
        checkpointer = ModelCheckpoint(monitor=self.monitor, filepath=self.weights_dir, 
                                       verbose=self.verbose, save_best_only=self.save_best_only, 
                                       save_weight_only=self.save_weight_only)
        reduce_lr = ReduceLROnPlateau(monitor=self.monitor, patience=self.patience, 
                                      factor=self.factor, mode=self.mode, verbose=self.verbose, 
                                      min_lr=self.min_lr)
        es = EarlyStopping(monitor=self.monitor, mode=self.mode, verbose=self.verbose, 
                           patience=self.es_patience)
        model = self.build()
        model.summary()
        
        if self.class_balancing: # compute_class_weight if class_balancing is True
            class_weight  = compute_class_weight(y_train)
            self.loss[-1] = SparseCategoricalCrossentropy(class_weight=class_weight)
            
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics, loss_weights=self.loss_weights)

        model.fit(x=X_train, y=[y_train, y_train],
                    batch_size=self.batch_size, shuffle=self.shuffle, 
                    epochs=self.epochs, validation_data=(X_val, [y_val,y_val]),
                    callbacks=[checkpointer,csv_logger,reduce_lr,es, time_callback])
        
    def predict(self, X_test, y_test):  

        if X_test.ndim != 4:
            raise Exception('ValueError: `X_test` is incompatible: expected ndim=4, found ndim={}'.format(X_test.ndim))

        model = self.build()
        model.load_weights(self.weights_dir)
        model.compile(optimizer=self.optimizer, loss=self.loss, metrics=self.metrics, loss_weights=self.loss_weights)

        start = time.time()
        y_pred_trip, y_pred_clf = model.predict(X_test)
        end = time.time()
        loss, trip_loss, classifier_loss, trip_acc, classifier_acc  = model.evaluate(x=X_test,
                                                                                     y=[y_test,y_test],
                                                                                     batch_size=self.batch_size, 
                                                                                     verbose=self.verbose)
        
        y_pred_argm = np.argmax(y_pred_clf, axis=1)
        f1 = f1_score(y_test, y_pred_argm, average=self.f1_average)
        print("F1-score is computed based on {}".format(self.f1_average))
        print('(loss: {}, accuracy: {})'.format(loss, classifier_acc))
        print(classification_report(y_test, y_pred_argm))
        evaluation = {'loss': loss, 
                      'triplet_loss': trip_loss, 
                      'classifier_loss': classifier_loss, 
                    'accuracy': classifier_acc, 
                    'f1-score': f1, 
                    'prediction_time': end-start}
        Y = {'y_true': y_test, 
             'y_pred': y_pred_argm}

        return Y, evaluation


import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="min2net",
    version="1.0.1",
    author="INTERFACES",
    author_email="IoBT.VISTEC@gmail.com",
    description="MIN2Net: End-to-End Multi-Task Learning for Subject-Independent Motor Imagery EEG Classification",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://MIN2Net.github.io",
    download_url="https://github.com/IoBT-VISTEC/MIN2Net/releases",
    project_urls={
        "Bug Tracker": "https://github.com/IoBT-VISTEC/MIN2Net/issues",
        "Documentation": "https://MIN2Net.github.io",
        "Source Code": "https://github.com/IoBT-VISTEC/MIN2Net",
    },
    license="Apache Software License",
    keywords=[
        "Brain-computer Interfaces"
        "BCI", 
        "Motor Imagery",
        "MI", 
        "Multi-task Learning",
        "Deep Metric Learning",
        "DML", 
        "Autoencoder",
        "AE",
        "EEG Classifier"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires = [
        "setuptools>=42",
        "wheel>=0.37.0",
        # "tensorflow-gpu==2.2.0", not support tensorflow-gpu via pip
        "tensorflow-addons==0.9.1",
        "scikit-learn>=0.24.1",
        "wget>=3.2"
    ],
    packages=setuptools.find_packages(),
    python_requires='>=3.6',

)



import numpy as np
import scipy.io as sio
from min2net.utils import resampling
from min2net.preprocessing.config import CONSTANT
from sklearn.model_selection import train_test_split
CONSTANT = CONSTANT['OpenBMI']

def read_raw(PATH, session, subject, num_class, id_ch_selected):
    mat_file_name = PATH + '/sess'+str(session).zfill(2)+'_subj'+str(subject).zfill(2)+'_EEG_MI.mat'
    mat = sio.loadmat(mat_file_name)
    print('This is data from: ', mat_file_name)
    if num_class == 2:
        raw_train_data = mat['EEG_MI_train'][0]['smt'][0]
        raw_train_data = (np.swapaxes(raw_train_data, 0, 2))[id_ch_selected]
        raw_train_data = np.swapaxes(raw_train_data, 0, 1)
        print('raw_train_data_shape:', raw_train_data.shape)
        raw_test_data = mat['EEG_MI_test'][0]['smt'][0]
        raw_test_data = np.swapaxes(raw_test_data, 0, 2)[id_ch_selected]
        raw_test_data = np.swapaxes(raw_test_data, 0, 1)
        print('raw_test_data_shape:', raw_test_data.shape)
        label_train_data = mat['EEG_MI_train'][0]['y_dec'][0][0]-1
        label_test_data = mat['EEG_MI_test'][0]['y_dec'][0][0]-1
        return raw_train_data, label_train_data, raw_test_data, label_test_data

    elif num_class == 3:
        raw_train_data = __segment_data(mat, type_data = 'train')
        raw_train_data = np.take(raw_train_data, id_ch_selected, axis=2)
        raw_train_data = np.swapaxes(raw_train_data, 1, 2)
        print('raw_train_data_shape:', raw_train_data.shape)
        raw_test_data = __segment_data(mat, type_data = 'test')
        raw_test_data = np.take(raw_test_data, id_ch_selected, axis=2)
        raw_test_data = np.swapaxes(raw_test_data, 1, 2)
        print('raw_test_data_shape:', raw_test_data.shape)
        label_train_data = mat['EEG_MI_train'][0]['y_dec'][0][0]-1
        label_test_data = mat['EEG_MI_test'][0]['y_dec'][0][0]-1
        return raw_train_data, label_train_data, raw_test_data, label_test_data
    
    elif num_class == "transitory_mi":
        raw_train_data = __segment_data_whole_period(mat, type_data = 'train')
        raw_train_data = np.take(raw_train_data, id_ch_selected, axis=2)
        raw_train_data = np.swapaxes(raw_train_data, 1, 2)
        print('raw_train_data_shape:', raw_train_data.shape)
        raw_test_data = __segment_data_whole_period(mat, type_data = 'test')
        raw_test_data = np.take(raw_test_data, id_ch_selected, axis=2)
        raw_test_data = np.swapaxes(raw_test_data, 1, 2)
        print('raw_test_data_shape:', raw_test_data.shape)
        label_train_data = mat['EEG_MI_train'][0]['y_dec'][0][0]-1
        label_test_data = mat['EEG_MI_test'][0]['y_dec'][0][0]-1
        return raw_train_data, label_train_data, raw_test_data, label_test_data

def load_crop_data(PATH, n_subjs, new_smp_freq, num_class, MI_len, id_chosen_chs, start_mi=None, stop_mi=None):
    if num_class == 2:
        print("Two-class MI data is downloading")
        orig_smp_freq = CONSTANT['orig_smp_freq'] # 1000
        n_trials = CONSTANT['n_trials_2_class'] # 100
        sessions = [1, 2]
        n_chs = len(id_chosen_chs)
        X_train, y_train = np.zeros((n_subjs,len(sessions),n_trials,n_chs,int(new_smp_freq* MI_len))), np.zeros((n_subjs,len(sessions),n_trials))
        X_test, y_test = np.zeros((n_subjs,len(sessions),n_trials,n_chs,int(new_smp_freq* MI_len))), np.zeros((n_subjs,len(sessions),n_trials))
        for id_sub, subject in enumerate(range(1, n_subjs+1)):
            for id_se, sess in enumerate(sessions):
                X_tr, y_tr, X_te, y_te = read_raw(PATH, sess, subject, num_class, id_chosen_chs)
                X_tr_resam = resampling(X_tr, new_smp_freq, MI_len)
                X_te_resam = resampling(X_te, new_smp_freq, MI_len)
                X_train[id_sub, id_se, :, :, :] = X_tr_resam
                X_test[id_sub, id_se, :, :, :] = X_te_resam
                y_train[id_sub, id_se, :] = y_tr
                y_test[id_sub, id_se, :] = y_te
        return X_train.reshape(n_subjs,-1,n_chs,int(new_smp_freq* MI_len)), y_train.reshape(n_subjs,-1), X_test.reshape(n_subjs,-1,n_chs,int(new_smp_freq* MI_len)), y_test.reshape(n_subjs,-1)
    elif num_class == 3:
        print("Three-class MI data is downloading")
        orig_smp_freq = CONSTANT['orig_smp_freq'] #1000
        n_trials = CONSTANT['n_trials_3_class'] #150
        sessions = [1, 2]
        n_chs = len(id_chosen_chs)
        MI_len = 4
        X_train, y_train = np.zeros((n_subjs,len(sessions),n_trials,n_chs,int(new_smp_freq* MI_len))), np.zeros((n_subjs,len(sessions),n_trials))
        X_test, y_test = np.zeros((n_subjs,len(sessions),n_trials,n_chs,int(new_smp_freq* MI_len))), np.zeros((n_subjs,len(sessions),n_trials))
        for id_sub, subject in enumerate(range(1, n_subjs+1)):
            for id_se, sess in enumerate(sessions):
                X_tr, y_tr, X_te, y_te = read_raw(PATH, sess, subject, num_class, id_chosen_chs)
                X_tr_addon, y_tr_addon = __add_on_resting(X_tr, y_tr, orig_smp_freq)
                X_te_addon, y_te_addon = __add_on_resting(X_te, y_te, orig_smp_freq)
                X_tr_resam = resampling(X_tr_addon, new_smp_freq, MI_len)
                X_te_resam = resampling(X_te_addon, new_smp_freq, MI_len)
                X_train[id_sub, id_se, :, :, :] = X_tr_resam
                X_test[id_sub, id_se, :, :, :] = X_te_resam
                y_train[id_sub, id_se, :] = y_tr_addon
                y_test[id_sub, id_se, :] = y_te_addon
        return X_train.reshape(n_subjs,-1,n_chs,int(new_smp_freq* MI_len)), y_train.reshape(n_subjs,-1), X_test.reshape(n_subjs,-1,n_chs,int(new_smp_freq* MI_len)), y_test.reshape(n_subjs,-1)
    
    elif num_class == "transitory_mi" and start_mi != None and stop_mi != None:
        print("Two-class transitory MI data is downloading with the time interval of {} s and {} s".format(start_mi, stop_mi))
        orig_smp_freq = CONSTANT['orig_smp_freq'] # 1000
        n_trials = CONSTANT['n_trials_2_class'] # 100
        sessions = [1, 2]
        n_chs = len(id_chosen_chs)
        MI_len = 4
        X_test, y_test = np.zeros((n_subjs,len(sessions),n_trials,n_chs,int(new_smp_freq* MI_len))), np.zeros((n_subjs,len(sessions),n_trials))
        for id_sub, subject in enumerate(range(1, n_subjs+1)):
            for id_se, sess in enumerate(sessions):
                X_tr, y_tr, X_te, y_te = read_raw(PATH, sess, subject, num_class, id_chosen_chs)
                X_te_transient, y_te_transient = __transitory_mi(X_te, y_te, orig_smp_freq, start_mi, stop_mi)
                X_te_resam = resampling(X_te_transient, new_smp_freq, MI_len)
                X_test[id_sub, id_se, :, :, :] = X_te_resam
                y_test[id_sub, id_se, :] = y_te_transient
        return X_test.reshape(n_subjs,-1,n_chs,int(new_smp_freq* MI_len)), y_test.reshape(n_subjs,-1)
                
def chanel_selection(sel_chs):
    orig_chs = CONSTANT['orig_chs']
    chs_id = []
    for name_ch in sel_chs:
        ch_id = np.where(np.array(orig_chs) == name_ch)[0][0]
        chs_id.append(ch_id)
        print('chosen_channel:', name_ch ,'---', 'Index_is:', ch_id)
    return chs_id

def __segment_data(mat_arr, type_data):
    data = mat_arr['EEG_MI_'+type_data][0]['x'][0]
    t = mat_arr['EEG_MI_'+type_data][0]['t'][0][0]
    low_cut = 0
    high_cut = CONSTANT['trial_len'] #8s
    orig_smp_freq = CONSTANT['orig_smp_freq'] #1000
    orig_n_chs = CONSTANT['n_chs'] #62
    n_trials = CONSTANT['n_trials_2_class'] #100
    data_seg = np.zeros((n_trials, high_cut*orig_smp_freq, orig_n_chs))
    # print('This pre-processing is for task {} low cut {} high cut {}'.format(task,low_cut,high_cut))
    for i in range(n_trials):
        start_pos = t[i]+(low_cut*orig_smp_freq)
        stop_pos = t[i]+(high_cut*orig_smp_freq)
        data_seg[i, :, :] = data[start_pos:stop_pos, :]
    return data_seg

def __segment_data_whole_period(mat_arr, type_data):
    data = mat_arr['EEG_MI_'+type_data][0]['x'][0]
    t = mat_arr['EEG_MI_'+type_data][0]['t'][0][0]
    low_cut = -3
    high_cut = CONSTANT['trial_len'] #8s
    data_len = 11
    orig_smp_freq = CONSTANT['orig_smp_freq'] #1000
    orig_n_chs = CONSTANT['n_chs'] #62
    n_trials = CONSTANT['n_trials_2_class'] #100
    data_seg = np.zeros((n_trials, data_len*orig_smp_freq, orig_n_chs))
    # print('This pre-processing is for task {} low cut {} high cut {}'.format(task,low_cut,high_cut))
    for i in range(n_trials):
        start_pos = t[i]+(low_cut*orig_smp_freq)
        stop_pos = t[i]+(high_cut*orig_smp_freq)
        # print("Debugg the selected period is:", (stop_pos-start_pos)/orig_smp_freq)
        data_seg[i, :, :] = data[start_pos:stop_pos, :]
    return data_seg

def __add_on_resting(X, y, smp_freq):
    print("MI Right, MI Left and Resting EEG Segmentation Process is being processed...")
    print("This data contains {} time ponts with sampling frequency of {} Hz.".format(X.shape[2], smp_freq))
    start_pos_mi = int(CONSTANT['MI']['start']*smp_freq) #0s
    stop_pos_mi = int(CONSTANT['MI']['stop']*smp_freq) #4s
    start_pos_rest = int(CONSTANT['MI']['stop']*smp_freq) #4s
    stop_pos_rest = int(CONSTANT['trial_len']*smp_freq) #8s
    index_class1 = np.where(y == 0)[0]
    index_class2 = np.where(y == 1)[0]
    X_class1, y_class1 = X[index_class1], y[index_class1]
    X_class2, y_class2 = X[index_class2], y[index_class2]
    # Split data into resting and MI signals
    X_mi_class1 = X_class1[:,:,start_pos_mi:stop_pos_mi]
    X_mi_class2 = X_class2[:,:,start_pos_mi:stop_pos_mi]
    X_rest_class1 = X_class1[:,:,start_pos_rest:stop_pos_rest]
    X_rest_class2 = X_class2[:,:,start_pos_rest:stop_pos_rest]
    # Choose a half of resting data to keep balancing the number of classes in our data
    X_rest_class1_50per,_ ,_ ,_= train_test_split(X_rest_class1, y_class1, random_state=42, test_size=0.5)
    X_rest_class2_50per,_ ,_ ,_= train_test_split(X_rest_class2, y_class2, random_state=42, test_size=0.5)
    X_rest_all = np.concatenate((X_rest_class1_50per, X_rest_class2_50per), axis=0)
    # Build class for resting data
    y_rest_all = np.full(X_rest_all.shape[0], 2)
    # Combine all classes again
    X_new_all = np.concatenate((X_mi_class1, X_mi_class2, X_rest_all), axis=0)
    y_new_all = np.concatenate((y_class1, y_class2, y_rest_all), axis=0)
    return X_new_all, y_new_all

def __transitory_mi(X, y, smp_freq, start, stop):
    print("MI Right ang MI Left EEG including transitory period is being processed...")
    print("This data contains {} time ponts with sampling frequency of {} Hz.".format(X.shape[2], smp_freq))
    start_pos_mi = int(start*smp_freq)
    stop_pos_mi = int(stop*smp_freq)
    # Segment needed MI period
    X_mi = X[:,:,start_pos_mi:stop_pos_mi]
    return X_mi, y
